<?php

/***************************************************************************
 *
 * Copyright (c) by Andreas Unterkircher, unki@netshadow.at
 * All rights reserved
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

require_once "phpfspot_cfg.php";
require_once "phpfspot_db.php";
require_once "phpfspot_tmpl.php";

class PHPFSPOT {

   var $cfg;
   var $db;
   var $cfg_db;
   var $tmpl;
   var $tags;
   var $avail_tags;

   /**
    * class constructor
    *
    * this function will be called on class construct
    * and will check requirements, loads configuration,
    * open databases and start the user session
    */
   public function __construct()
   {
      /* Check necessary requirements */
      if(!$this->checkRequirements()) {
         exit(1);
      }

      $this->cfg = new PHPFSPOT_CFG;

      $this->db  = new PHPFSPOT_DB(&$this, $this->cfg->fspot_db);
      
      if(!is_writeable(dirname($this->cfg->phpfspot_db))) {
         print dirname($this->cfg->phpfspot_db) .": directory is not writeable!";
         exit(1);
      }
         
      $this->cfg_db = new PHPFSPOT_DB(&$this, $this->cfg->phpfspot_db);
      $this->check_config_table();

      $this->tmpl = new PHPFSPOT_TMPL($this);

      $this->get_tags();

      session_start();

      if(!isset($_SESSION['tag_condition']))
         $_SESSION['tag_condition'] = 'or';

      if(!isset($_SESSION['searchfor']))
         $_SESSION['searchfor'] = '';

      // if begin_with is still set but rows_per_page is now 0, unset it
      if(isset($_SESSION['begin_with']) && $this->cfg->rows_per_page == 0)
         unset($_SESSION['begin_with']);

   } // __construct()

   public function __destruct()
   {

   } // __destruct()

   /**
    * show - generate html output
    *
    * this function can be called after the constructor has
    * prepared everyhing. it will load the index.tpl smarty
    * template. if necessary it will registere pre-selects
    * (photo index, photo, tag search, date search) into
    * users session.
    */
   public function show()
   {
      $this->tmpl->assign('searchfor', $_SESSION['searchfor']);
      $this->tmpl->assign('page_title', $this->cfg->page_title);
      $this->tmpl->assign('current_condition', $_SESSION['tag_condition']);

      $_SESSION['start_action'] = $_GET['mode'];

      switch($_GET['mode']) {
         case 'showpi':
            if(isset($_GET['tags'])) {
               $_SESSION['selected_tags'] = split(',', $_GET['tags']);
            }
            if(isset($_GET['from_date'])) {
               $_SESSION['from_date'] = $_GET['from_date'];
            }
            if(isset($_GET['to_date'])) {
               $_SESSION['to_date'] = $_GET['to_date'];
            }
            break;
         case 'showp':
            if(isset($_GET['tags'])) {
               $_SESSION['selected_tags'] = split(',', $_GET['tags']);
               $_SESSION['start_action'] = 'showp';
            }
            if(isset($_GET['id'])) {
               $_SESSION['current_photo'] = $_GET['id'];
               $_SESSION['start_action'] = 'showp';
            }
            if(isset($_GET['from_date'])) {
               $_SESSION['from_date'] = $_GET['from_date'];
            }
            if(isset($_GET['to_date'])) {
               $_SESSION['to_date'] = $_GET['to_date'];
            }
            break;
         case 'export':
            $this->tmpl->show("export.tpl");
            return;
            break;
      }

      $this->tmpl->assign('from_date', $this->get_calendar('from'));
      $this->tmpl->assign('to_date', $this->get_calendar('to'));
      $this->tmpl->assign('content_page', 'welcome.tpl');
      $this->tmpl->show("index.tpl");


   } // show()

   /**
    * get_tags - grab all tags of f-spot's database
    *
    * this function will get all available tags from
    * the f-spot database and store them within two
    * arrays within this clase for later usage. in
    * fact, if the user requests (hide_tags) it will
    * opt-out some of them.
    *
    * this function is getting called once by show()
    */
   private function get_tags()
   {
      $this->avail_tags = Array();
      $count = 0;
   
      $result = $this->db->db_query("
         SELECT id,name
         FROM tags
         ORDER BY sort_priority ASC
      ");
      
      while($row = $this->db->db_fetch_object($result)) {

         $tag_id = $row['id'];
         $tag_name = $row['name'];

         /* check if config requests to ignore this tag */
         if(in_array($row['name'], $this->cfg->hide_tags))
            continue;

         $this->tags[$tag_id] = $tag_name; 
         $this->avail_tags[$count] = $tag_id;

         $count++;

      }

   } // get_tags()

   /** 
    * extract all photo details
    * 
    * retrieve all available details from f-spot's
    * database and return them as object
    */
   public function get_photo_details($idx)
   {
      $result = $this->db->db_query("
         SELECT *
         FROM photos
         WHERE id='". $idx ."'
      ");
      
      return $this->db->db_fetch_object($result);

   } // get_photo_details

   /**
    * returns aligned photo names 
    *
    * this function returns aligned (length) names for
    * an specific photo. If the length of the name exceeds
    * $limit the name will be shrinked (...)
    */
   public function getPhotoName($idx, $limit = 0)
   {
      if($details = $this->get_photo_details($idx)) {
         $name = $details['name'];
         if($limit != 0 && strlen($name) > $limit) {
            $name = substr($name, 0, $limit-5) ."...". substr($name, -($limit-5));
         }
         return $name;
      }

   } // getPhotoName()

   /**
    * translate f-spoth photo path
    * 
    * as the full-qualified path recorded in the f-spot database
    * is usally not the same as on the webserver, this function
    * will replace the path with that one specified in the cfg
    */
   public function translate_path($path, $width = 0)
   {  
      return str_replace($this->cfg->path_replace_from, $this->cfg->path_replace_to, $path);

   } // translate_path

   /**
    * control HTML ouput for a single photo
    *
    * this function provides all the necessary information
    * for the single photo template.
    */
   public function showPhoto($photo)
   {
      /* get all photos from the current photo selection */
      $all_photos = $this->getPhotoSelection();
      $count = count($all_photos);

      for($i = 0; $i < $count; $i++) {
         
         // $get_next will be set, when the photo which has to
         // be displayed has been found - this means that the
         // next available is in fact the NEXT image (for the
         // navigation icons) 
         if(isset($get_next)) {
            $next_img = $all_photos[$i];
            break;
         }

         /* the next photo is our NEXT photo */
         if($all_photos[$i] == $photo) {
            $get_next = 1;
         }
         else {
            $previous_img = $all_photos[$i];
         }

         if($photo == $all_photos[$i]) {
               $current = $i;
         }
      }

      $details = $this->get_photo_details($photo);

      if(!$details) {
         print "error";
         return;
      }

      $orig_path = $this->translate_path($details['directory_path']) ."/". $details['name'];
      $thumb_path = $this->cfg->base_path ."/thumbs/". $this->cfg->photo_width ."_". $this->getMD5($photo);

      if(!file_exists($orig_path)) {
         $this->_warning("Photo ". $orig_path ." does not exist!<br />\n");
      }

      if(!is_readable($orig_path)) {
         $this->_warning("Photo ". $orig_path ." is not readable for user ". $this->getuid() ."<br />\n");
      }

      /* If the thumbnail doesn't exist yet, try to create it */
      if(!file_exists($thumb_path)) {
         $this->gen_thumb($photo, true);
         $thumb_path = $this->cfg->base_path ."/thumbs/". $this->cfg->photo_width ."_". $this->getMD5($photo);
      }

      /* get f-spot database meta information */
      $meta = $this->get_meta_informations($orig_path);

      /* If EXIF data are available, use them */
      if(isset($meta['ExifImageWidth'])) {
         $meta_res = $meta['ExifImageWidth'] ."x". $meta['ExifImageLength'];
      } else {
         $info = getimagesize($orig_path);
         $meta_res = $info[0] ."x". $info[1]; 
      }

      $meta_date = isset($meta['FileDateTime']) ? strftime("%a %x %X", $meta['FileDateTime']) : "n/a";
      $meta_make = isset($meta['Make']) ? $meta['Make'] ." / ". $meta['Model'] : "n/a";
      $meta_size = isset($meta['FileSize']) ? round($meta['FileSize']/1024, 1) ."kbyte" : "n/a";

      $extern_link = "index.php?mode=showp&id=". $photo;
      $current_tags = $this->getCurrentTags();
      if($current_tags != "") {
         $extern_link.= "&tags=". $current_tags;
      }
      if(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
         $extern_link.= "&from_date=". $_SESSION['from_date'] ."&to_date=". $_SESSION['to_date'];
      }

      $this->tmpl->assign('extern_link', $extern_link);

      if(file_exists($thumb_path)) {

         $info = getimagesize($thumb_path);

         $this->tmpl->assign('description', $details['description']);
         $this->tmpl->assign('image_name', $details['name']);

         $this->tmpl->assign('width', $info[0]);
         $this->tmpl->assign('height', $info[1]);
         $this->tmpl->assign('ExifMadeOn', $meta_date);
         $this->tmpl->assign('ExifMadeWith', $meta_make);
         $this->tmpl->assign('ExifOrigResolution', $meta_res);
         $this->tmpl->assign('ExifFileSize', $meta_size);
    
         $this->tmpl->assign('image_url', 'phpfspot_img.php?idx='. $photo ."&amp;width=". $this->cfg->photo_width);
         $this->tmpl->assign('image_url_full', 'phpfspot_img.php?idx='. $photo);

         $this->tmpl->assign('tags', $this->get_photo_tags($photo));
         $this->tmpl->assign('current', $current);
      }
      else {
         $this->_warning("Can't open file ". $thumb_path ."\n");
         return;
      }

      if($previous_img) {
         $this->tmpl->assign('previous_url', "javascript:showImage(". $previous_img .");");
         $this->tmpl->assign('prev_img', $previous_img);
      }

      if($next_img) {
         $this->tmpl->assign('next_url', "javascript:showImage(". $next_img .");");
         $this->tmpl->assign('next_img', $next_img);
      }
      $this->tmpl->assign('mini_width', $this->cfg->mini_width);

      $this->tmpl->show("single_photo.tpl");

   } // showPhoto()

   /**
    * all available tags and tag cloud
    *
    * this function outputs all available tags (time ordered)
    * and in addition output them as tag cloud (tags which have
    * many photos will appears more then others)
    */
   public function getAvailableTags()
   {
      $output = "";

      $result = $this->db->db_query("
         SELECT tag_id as id, count(tag_id) as quantity
         FROM photo_tags
         INNER JOIN tags t
            ON t.id = tag_id
         GROUP BY tag_id
         ORDER BY t.name ASC
      ");

      $tags = Array();

      while($row = $this->db->db_fetch_object($result)) {
         $tags[$row['id']] = $row['quantity'];
      }

      // change these font sizes if you will
      $max_size = 125; // max font size in %
      $min_size = 75; // min font size in %

      // get the largest and smallest array values
      $max_qty = max(array_values($tags));
      $min_qty = min(array_values($tags));

      // find the range of values
      $spread = $max_qty - $min_qty;
      if (0 == $spread) { // we don't want to divide by zero
         $spread = 1;
      }

      // determine the font-size increment
      // this is the increase per tag quantity (times used)
      $step = ($max_size - $min_size)/($spread);

      // loop through our tag array
      foreach ($tags as $key => $value) {

         if(isset($_SESSION['selected_tags']) && in_array($key, $_SESSION['selected_tags']))
            continue;

          // calculate CSS font-size
          // find the $value in excess of $min_qty
          // multiply by the font-size increment ($size)
          // and add the $min_size set above
         $size = $min_size + (($value - $min_qty) * $step);
          // uncomment if you want sizes in whole %:
         $size = ceil($size);

         $output.= "<a href=\"javascript:Tags('add', ". $key .");\" class=\"tag\" style=\"font-size: ". $size ."%;\">". $this->tags[$key] ."</a>, ";

      }

      $output = substr($output, 0, strlen($output)-2);
      print $output;

   } // getAvailableTags()

   /**
    * output all selected tags
    *
    * this function output all tags which have been selected
    * by the user. the selected tags are stored in the 
    * session-variable $_SESSION['selected_tags']
    */
   public function getSelectedTags()
   {
      $output = "";
      foreach($this->avail_tags as $tag)
      {
         // return all selected tags
         if(isset($_SESSION['selected_tags']) && in_array($tag, $_SESSION['selected_tags'])) {
            $output.= "<a href=\"javascript:Tags('del', ". $tag .");\" class=\"tag\">". $this->tags[$tag] ."</a>, ";
         }
      }

      $output = substr($output, 0, strlen($output)-2);
      print $output;

   } // getSelectedTags()

   /**
    * add tag to users session variable
    *
    * this function will add the specified to users current
    * tag selection. if a date search has been made before
    * it will be now cleared
    */
   public function addTag($tag)
   {
      if(!isset($_SESSION['selected_tags']))
         $_SESSION['selected_tags'] = Array();

      if(!in_array($tag, $_SESSION['selected_tags']))
         array_push($_SESSION['selected_tags'], $tag);
   
   } // addTag()

   /**
    * remove tag to users session variable
    *
    * this function removes the specified tag from
    * users current tag selection
    */
   public function delTag($tag)
   {
      if(isset($_SESSION['selected_tags'])) {
         $key = array_search($tag, $_SESSION['selected_tags']);
         unset($_SESSION['selected_tags'][$key]);
         sort($_SESSION['selected_tags']);
      }

   } // delTag()

   /**
    * reset tag selection
    *
    * if there is any tag selection, it will be
    * deleted now
    */
   public function resetTags()
   {
      if(isset($_SESSION['selected_tags']))
         unset($_SESSION['selected_tags']);

   } // resetTags()

   /**
    * reset single photo
    *
    * if a specific photo was requested (external link)
    * unset the session variable now
    */
   public function resetPhotoView()
   {
      if(isset($_SESSION['current_photo']))
         unset($_SESSION['current_photo']);

   } // resetPhotoView();

   /**
    * reset tag search
    *
    * if any tag search has taken place, reset
    * it now
    */
   public function resetTagSearch()
   {
      if(isset($_SESSION['searchfor']))
         unset($_SESSION['searchfor']);

   } // resetTagSearch()

    /**
    * reset date search
    *
    * if any date search has taken place, reset
    * it now
    */
   public function resetDateSearch()
   {
      if(isset($_SESSION['from_date']))
         unset($_SESSION['from_date']);
      if(isset($_SESSION['to_date']))
         unset($_SESSION['to_date']);

   } // resetDateSearch();

   /**
    * return all photo according selection
    *
    * this function returns all photos based on
    * the tag-selection, tag- or date-search.
    * the tag-search also has to take care of AND
    * and OR conjunctions
    */
   public function getPhotoSelection()
   {  
      $matched_photos = Array();

      if(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
         $from_date = strtotime($_SESSION['from_date']);
         $to_date = strtotime($_SESSION['to_date']);
         $additional_where_cond = "
               p.time>='". $from_date ."'
            AND
               p.time<='". $to_date ."'
         ";
      } 

      /* return a search result */
      if(isset($_SESSION['searchfor']) && $_SESSION['searchfor'] != '') {
         $query_str = "
            SELECT DISTINCT photo_id
               FROM photo_tags pt
            INNER JOIN photos p
               ON p.id=pt.photo_id
            INNER JOIN tags t
               ON pt.tag_id=t.id
            WHERE t.name LIKE '%". $_SESSION['searchfor'] ."%'";

         if(isset($additional_where_cond))
            $query_str.= "AND ". $additional_where_cond ." ";
         $query_str.= "
            ORDER BY p.time ASC
         ";
         $result = $this->db->db_query($query_str);
         while($row = $this->db->db_fetch_object($result)) {
            array_push($matched_photos, $row['photo_id']);
         }
         return $matched_photos;
      }

      /* return according the selected tags */
      if(isset($_SESSION['selected_tags']) && !empty($_SESSION['selected_tags'])) {
         $selected = "";
         foreach($_SESSION['selected_tags'] as $tag)
            $selected.= $tag .",";
         $selected = substr($selected, 0, strlen($selected)-1);

         if($_SESSION['tag_condition'] == 'or') {
            $query_str = "
               SELECT DISTINCT photo_id
                  FROM photo_tags pt
               INNER JOIN photos p
                  ON p.id=pt.photo_id
               WHERE pt.tag_id IN (". $selected .")
            ";
            if(isset($additional_where_cond)) 
               $query_str.= "AND ". $additional_where_cond ." ";

            $query_str.= "
               ORDER BY p.time ASC
            ";
         }
         elseif($_SESSION['tag_condition'] == 'and') {

            if(count($_SESSION['selected_tags']) >= 32) {
               print "A SQLite limit of 32 tables within a JOIN SELECT avoids to<br />\n";
               print "evaluate your tag selection. Please remove some tags from your selection.\n";
               return Array();
            } 

            /* Join together a table looking like

               pt1.photo_id pt1.tag_id pt2.photo_id pt2.tag_id ...

               so the query can quickly return all images matching the
               selected tags in an AND condition

            */

            $query_str = "
               SELECT DISTINCT pt1.photo_id
                  FROM photo_tags pt1
            ";

            for($i = 0; $i < count($_SESSION['selected_tags']); $i++) {
               $query_str.= "
                  INNER JOIN photo_tags pt". ($i+2) ."
                     ON pt1.photo_id=pt". ($i+2) .".photo_id
               ";
            }
            $query_str.= "WHERE pt1.tag_id=". $_SESSION['selected_tags'][0];
            for($i = 1; $i < count($_SESSION['selected_tags']); $i++) {
               $query_str.= "
                  AND pt". ($i+1) .".tag_id=". $_SESSION['selected_tags'][$i] ."
               "; 
            }
            if(isset($additional_where_cond)) 
               $query_str.= "AND ". $additional_where_cond;
         }

         $result = $this->db->db_query($query_str);
         while($row = $this->db->db_fetch_object($result)) {
            array_push($matched_photos, $row['photo_id']);
         }
         return $matched_photos;
      }

      /* return all available photos */
      $query_str = "
         SELECT DISTINCT photo_id
            FROM photo_tags pt
         INNER JOIN photos p
            ON p.id=pt.photo_id
      ";
      if(isset($additional_where_cond)) 
         $query_str.= "WHERE ". $additional_where_cond ." ";
      $query_str.= "
         ORDER BY p.time ASC
      ";
      $result = $this->db->db_query($query_str);
      while($row = $this->db->db_fetch_object($result)) {
         array_push($matched_photos, $row['photo_id']);
      }
      return $matched_photos;

   } // getPhotoSelection()

    /**
    * control HTML ouput for photo index
    *
    * this function provides all the necessary information
    * for the photo index template.
    */
   public function showPhotoIndex()
   {
      $photos = $this->getPhotoSelection();

      $count = count($photos);

      if(isset($_SESSION['begin_with']) && $_SESSION['begin_with'] != "")
         $anchor = $_SESSION['begin_with'];

      if(!isset($this->cfg->rows_per_page) || $this->cfg->rows_per_page == 0) {

         $begin_with = 0;
         $end_with = $count;

      }
      elseif($this->cfg->rows_per_page > 0) {

         if(!$_SESSION['begin_with'] || $_SESSION['begin_with'] == 0)
            $begin_with = 0;
         else {

            $begin_with = $_SESSION['begin_with'];

            // verify $begin_with - perhaps the thumbs-per-rows or
            // rows-per-page variables have changed or the jump back
            // from a photo wasn't exact - so calculate the real new
            // starting point
            $multiplicator = $this->cfg->rows_per_page * $this->cfg->thumbs_per_row;
            for($i = 0; $i <= $count; $i+=$multiplicator) {
               if($begin_with >= $i && $begin_with < $i+$multiplicator) {
                  $begin_with = $i;
                  break;
               }
            }
         }

         $end_with = $begin_with + ($this->cfg->rows_per_page * $this->cfg->thumbs_per_row);
      }

   
      $rows = 0;
      $cols = 0;
      $images[$rows] = Array();
      $img_height[$rows] = Array();
      $img_width[$rows] = Array();
      $img_id[$rows] = Array();
      $img_name[$rows] = Array();
      $img_title = Array();

      for($i = $begin_with; $i < $end_with; $i++) {

         $images[$rows][$cols] = $photos[$i];
         $img_id[$rows][$cols] = $i;
         $img_name[$rows][$cols] = htmlspecialchars($this->getPhotoName($photos[$i], 15));
         $img_title[$rows][$cols] = "Click to view photo ". htmlspecialchars($this->getPhotoName($photos[$i], 0));

         $thumb_path = $this->cfg->base_path ."/thumbs/". $this->cfg->thumb_width ."_". $this->getMD5($photos[$i]);

         if(file_exists($thumb_path)) {
            $info = getimagesize($thumb_path); 
            $img_width[$rows][$cols] = $info[0];
            $img_height[$rows][$cols] = $info[1];
         }

         if($cols == $this->cfg->thumbs_per_row-1) {
            $cols = 0;
            $rows++;
            $images[$rows] = Array();
            $img_width[$rows] = Array();
            $img_height[$rows] = Array();
         }
         else {
            $cols++;
         }
      } 

      // +1 for for smarty's selection iteration
      $rows++;

      if(isset($_SESSION['searchfor']) && $_SESSION['searchfor'] != '')
         $this->tmpl->assign('searchfor', $_SESSION['searchfor']);

      if(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
         $this->tmpl->assign('from_date', $_SESSION['from_date']);
         $this->tmpl->assign('to_date', $_SESSION['to_date']);
      }

      if(isset($_SESSION['selected_tags']) && !empty($_SESSION['selected_tags'])) {
         $this->tmpl->assign('tag_result', 1);
      }

      /* do we have to display the page selector ? */
      if($this->cfg->rows_per_page != 0) {
      
         /* calculate the page switchers */
         $previous_start = $begin_with - ($this->cfg->rows_per_page * $this->cfg->thumbs_per_row);
         $next_start = $begin_with + ($this->cfg->rows_per_page * $this->cfg->thumbs_per_row);

         if($begin_with != 0) 
            $this->tmpl->assign("previous_url", "javascript:showPhotoIndex(". $previous_start .");"); 
         if($end_with < $count)
            $this->tmpl->assign("next_url", "javascript:showPhotoIndex(". $next_start .");"); 

         $photo_per_page  = $this->cfg->rows_per_page * $this->cfg->thumbs_per_row;
         $last_page = ceil($count / $photo_per_page);

         /* get the current selected page */
         if($begin_with == 0) {
            $current_page = 1;
         } else {
            $current_page = 0;
            for($i = $begin_with; $i >= 0; $i-=$photo_per_page) {
               $current_page++;
            }
         } 

         $dotdot_made = 0;

         for($i = 1; $i <= $last_page; $i++) {

            if($current_page == $i)
               $style = "style=\"font-size: 125%;\"";
            elseif($current_page-1 == $i || $current_page+1 == $i)
               $style = "style=\"font-size: 105%;\"";
            elseif(($current_page-5 >= $i) && ($i != 1) ||
               ($current_page+5 <= $i) && ($i != $last_page))
               $style = "style=\"font-size: 75%;\"";
            else
               $style = "";

            $select = "<a href=\"javascript:showPhotoIndex(". (($i*$photo_per_page)-$photo_per_page) .");\"";
               if($style != "")
                  $select.= $style;
            $select.= ">". $i ."</a>&nbsp;";

            // until 9 pages we show the selector from 1-9
            if($last_page <= 9) {
               $page_select.= $select;
               continue;
            } else {
               if($i == 1 /* first page */ || 
                  $i == $last_page /* last page */ ||
                  $i == $current_page /* current page */ ||
                  $i == ceil($last_page * 0.25) /* first quater */ ||
                  $i == ceil($last_page * 0.5) /* half */ ||
                  $i == ceil($last_page * 0.75) /* third quater */ ||
                  (in_array($i, array(1,2,3,4,5,6)) && $current_page <= 4) /* the first 6 */ ||
                  (in_array($i, array($last_page, $last_page-1, $last_page-2, $last_page-3, $last_page-4, $last_page-5)) && $current_page >= $last_page-4) /* the last 6 */ ||
                  $i == $current_page-3 || $i == $current_page-2 || $i == $current_page-1 /* three before */ ||
                  $i == $current_page+3 || $i == $current_page+2 || $i == $current_page+1 /* three after */) {

                  $page_select.= $select;
                  $dotdot_made = 0;
                  continue;

               }
            }

            if(!$dotdot_made) {
               $page_select.= ".........&nbsp;";
               $dotdot_made = 1;
            }
         }

         /* only show the page selector if we have more then one page */
         if($last_page > 1)
            $this->tmpl->assign('page_selector', $page_select);
      }

      
      $current_tags = $this->getCurrentTags();
      $extern_link = "index.php?mode=showpi";
      if($current_tags != "") {
         $extern_link.= "&tags=". $current_tags;
      }
      if(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
         $extern_link.= "&from_date=". $_SESSION['from_date'] ."&to_date=". $_SESSION['to_date'];
      }

      $export_link = "index.php?mode=export";

      $this->tmpl->assign('extern_link', $extern_link);
      $this->tmpl->assign('export_link', $export_link);
      $this->tmpl->assign('count', $count);
      $this->tmpl->assign('width', $this->cfg->thumb_width);
      $this->tmpl->assign('images', $images);
      $this->tmpl->assign('img_width', $img_width);
      $this->tmpl->assign('img_height', $img_height);
      $this->tmpl->assign('img_id', $img_id);
      $this->tmpl->assign('img_name', $img_name);
      $this->tmpl->assign('img_title', $img_title);
      $this->tmpl->assign('rows', $rows);
      $this->tmpl->assign('columns', $this->cfg->thumbs_per_row);

      $this->tmpl->show("photo_index.tpl");

      if(isset($anchor))
         print "<script language=\"JavaScript\">self.location.hash = '#image". $anchor ."';</script>\n";

   } // showPhotoIndex()

   /**
    * show credit template
    */
   public function showCredits()
   {
      $this->tmpl->assign('version', $this->cfg->version);
      $this->tmpl->assign('product', $this->cfg->product);
      $this->tmpl->show("credits.tpl");

   } // showCredits()

   /**
    * create_thumbnails for the requested width
    *
    * this function creates image thumbnails of $orig_image
    * stored as $thumb_image. It will check if the image is
    * in a supported format, if necessary rotate the image
    * (based on EXIF orientation meta headers) and re-sizing.
    */
   public function create_thumbnail($orig_image, $thumb_image, $width)
   {  
      if(!file_exists($orig_image)) {
         return false;
      }

      $details = getimagesize($orig_image);
      
      /* check if original photo is a support image type */
      if(!$this->checkifImageSupported($details['mime']))
         return false;

      $meta = $this->get_meta_informations($orig_image);

      $rotate = 0;
      $flip = false;

      switch($meta['Orientation']) {

         case 1: /* top, left */
            $rotate = 0; $flip = false; break;
         case 2: /* top, right */
            $rotate = 0; $flip = true; break;
         case 3: /* bottom, left */
            $rotate = 180; $flip = false; break;
         case 4: /* bottom, right */
            $rotate = 180; $flip = true; break;
         case 5: /* left side, top */
            $rotate = 90; $flip = true; break;
         case 6: /* right side, top */
            $rotate = 90; $flip = false; break;
         case 7: /* left side, bottom */
            $rotate = 270; $flip = true; break;
         case 8: /* right side, bottom */
            $rotate = 270; $flip = false; break;
      }

      $src_img = @imagecreatefromjpeg($orig_image);

      if(!$src_img) {
         print "Can't load image from ". $orig_image ."\n";
         return false;
      }

      /* grabs the height and width */
      $cur_width = imagesx($src_img);
      $cur_height = imagesy($src_img);

      // If requested width is more then the actual image width,
      // do not generate a thumbnail, instead safe the original
      // as thumbnail but with lower quality

      if($width >= $cur_width) {
         $result = imagejpeg($src_img, $thumb_image, 75);
         imagedestroy($src_img);
         return true;
      }

      // If the image will be rotate because EXIF orientation said so
      // 'virtually rotate' the image for further calculations
      if($rotate == 90 || $rotate == 270) {
         $tmp = $cur_width;
         $cur_width = $cur_height;
         $cur_height = $tmp;
      }

      /* calculates aspect ratio */
      $aspect_ratio = $cur_height / $cur_width;

      /* sets new size */
      if($aspect_ratio < 1) {
         $new_w = $width;
         $new_h = abs($new_w * $aspect_ratio);
      } else {
         /* 'virtually' rotate the image and calculate it's ratio */
         $tmp_w = $cur_height;
         $tmp_h = $cur_width;
         /* now get the ratio from the 'rotated' image */
         $tmp_ratio = $tmp_h/$tmp_w;
         /* now calculate the new dimensions */
         $tmp_w = $width;
         $tmp_h = abs($tmp_w * $tmp_ratio);

         // now that we know, how high they photo should be, if it
         // gets rotated, use this high to scale the image
         $new_h = $tmp_h;
         $new_w = abs($new_h / $aspect_ratio);

         // If the image will be rotate because EXIF orientation said so
         // now 'virtually rotate' back the image for the image manipulation
         if($rotate == 90 || $rotate == 270) {
            $tmp = $new_w;
            $new_w = $new_h;
            $new_h = $tmp;
         }
      }

      /* creates new image of that size */
      $dst_img = imagecreatetruecolor($new_w, $new_h);

      imagefill($dst_img, 0, 0, ImageColorAllocate($dst_img, 255, 255, 255));

      /* copies resized portion of original image into new image */
      imagecopyresampled($dst_img, $src_img, 0, 0, 0, 0, $new_w, $new_h, imagesx($src_img), imagesy($src_img));

      /* needs the image to be flipped horizontal? */
      if($flip) {
         print "(FLIP)";
         $image = $dst_img;
         for($x = 0; $x < $new_w; $x++) {
            imagecopy($dst_img, $image, $x, 0, $w - $x - 1, 0, 1, $h);
         }
      }

      if($rotate) {
         $this->_debug("(ROTATE)");
         $dst_img = $this->rotateImage($dst_img, $rotate);
      }

      /* write down new generated file */
      $result = imagejpeg($dst_img, $thumb_image, 75);

      /* free your mind */
      imagedestroy($dst_img);
      imagedestroy($src_img);

      if($result === false) {
         print "Can't write thumbnail ". $thumb_image ."\n";
         return false;
      }

      return true;

   } // create_thumbnail()

   /**
    * return all exif meta data from the file
    */
   public function get_meta_informations($file)
   {
      return exif_read_data($file);

   } // get_meta_informations()

   /**
    * create phpfspot own sqlite database
    *
    * this function creates phpfspots own sqlite database
    * if it does not exist yet. this own is used to store
    * some necessary informations (md5 sum's, ...).
    */
   public function check_config_table()
   {
      // if the config table doesn't exist yet, create it
      if(!$this->cfg_db->db_check_table_exists("images")) {
         $this->cfg_db->db_exec("
            CREATE TABLE images (
               img_idx int primary key,
               img_md5 varchar(32)
            )
            ");
      }

   } // check_config_table

   /**
    * Generates a thumbnail from photo idx
    *
    * This function will generate JPEG thumbnails from provided F-Spot photo
    * indizes.
    *
    * 1. Check if all thumbnail generations (width) are already in place and
    *    readable
    * 2. Check if the md5sum of the original file has changed
    * 3. Generate the thumbnails if needed
    */
   public function gen_thumb($idx = 0, $force = 0)
   {
      $error = 0;

      $resolutions = Array(
         $this->cfg->thumb_width,
         $this->cfg->photo_width,
         $this->cfg->mini_width,
      );

      /* get details from F-Spot's database */
      $details = $this->get_photo_details($idx);

      /* calculate file MD5 sum */
      $full_path = $this->translate_path($details['directory_path'])  ."/". $details['name'];

      if(!file_exists($full_path)) {
         $this->_warning("File ". $full_path ." does not exist\n");
         return;
      }

      if(!is_readable($full_path)) {
         $this->_warning("File ". $full_path ." is not readable for ". $this->getuid() ."\n");
         return;
      }

      $file_md5 = md5_file($full_path);

      $this->_debug("Image [". $idx ."] ". $details['name'] ." Thumbnails:");

      foreach($resolutions as $resolution) {

         $thumb_path = $this->cfg->base_path ."/thumbs/". $resolution ."_". $file_md5;

         /* if the thumbnail file doesn't exist, create it */
         if(!file_exists($thumb_path)) {

            $this->_debug(" ". $resolution ."px");
            if(!$this->create_thumbnail($full_path, $thumb_path, $resolution))
               $error = 1;
         }
         /* if the file hasn't changed there is no need to regen the thumb */
         elseif($file_md5 != $this->getMD5($idx) || $force) {

            $this->_debug(" ". $resolution ."px");
            if(!$this->create_thumbnail($full_path, $thumb_path, $resolution))
               $error = 1;

         }
      }

      /* set the new/changed MD5 sum for the current photo */
      if(!$error) {
         $this->setMD5($idx, $file_md5);
      }

      $this->_debug("\n");

   } // gen_thumb()

   /**
    * returns stored md5 sum for a specific photo
    *
    * this function queries the phpfspot database for a
    * stored MD5 checksum of the specified photo
    */
   public function getMD5($idx)
   {
      $result = $this->cfg_db->db_query("
         SELECT img_md5 
         FROM images
         WHERE img_idx='". $idx ."'
      ");

      if(!$result)
         return 0;

      $img = $this->cfg_db->db_fetch_object($result);
      return $img['img_md5'];
      
   } // getMD5()

   /**
    * set MD5 sum for the specific photo
    */
   private function setMD5($idx, $md5)
   {
      $result = $this->cfg_db->db_exec("
         REPLACE INTO images (img_idx, img_md5)
         VALUES ('". $idx ."', '". $md5 ."')
      ");

   } // setMD5()

   /**
    * store current tag condition
    *
    * this function stores the current tag condition
    * (AND or OR) in the users session variables
    */
   public function setTagCondition($mode)
   {
      $_SESSION['tag_condition'] = $mode;

   } // setTagCondition()

   /** 
    * invoke tag & date search 
    *
    * this function will return all matching tags and store
    * them in the session variable selected_tags. furthermore
    * it also handles the date search.
    * getPhotoSelection() will then only return the matching
    * photos.
    */
   public function startSearch($searchfor, $from, $to)
   {
      $_SESSION['searchfor'] = $searchfor;
      $_SESSION['from_date'] = $from;
      $_SESSION['to_date'] = $to;

      if($searchfor != "") {
         /* new search, reset the current selected tags */
         $_SESSION['selected_tags'] = Array();
         foreach($this->avail_tags as $tag) {
            if(preg_match('/'. $searchfor .'/i', $this->tags[$tag]))
               array_push($_SESSION['selected_tags'], $tag);
         }
      }
   } // startSearch()

   /**
    * rotate image
    *
    * this function rotates the image according the
    * specified angel.
    */
   private function rotateImage($img, $degrees)
   {
      if(function_exists("imagerotate")) {
         $img = imagerotate($img, $degrees, 0);
      } else {
         function imagerotate($src_img, $angle)
         {
            $src_x = imagesx($src_img);
            $src_y = imagesy($src_img);
            if ($angle == 180) {
               $dest_x = $src_x;
               $dest_y = $src_y;
            }
            elseif ($src_x <= $src_y) {
               $dest_x = $src_y;
               $dest_y = $src_x;
            }
            elseif ($src_x >= $src_y) {
               $dest_x = $src_y;
               $dest_y = $src_x;
            }
               
            $rotate=imagecreatetruecolor($dest_x,$dest_y);
            imagealphablending($rotate, false);
               
            switch ($angle) {
            
               case 90:
                  for ($y = 0; $y < ($src_y); $y++) {
                     for ($x = 0; $x < ($src_x); $x++) {
                        $color = imagecolorat($src_img, $x, $y);
                        imagesetpixel($rotate, $dest_x - $y - 1, $x, $color);
                     }
                  }
                  break;

               case 270:
                  for ($y = 0; $y < ($src_y); $y++) {
                     for ($x = 0; $x < ($src_x); $x++) {
                        $color = imagecolorat($src_img, $x, $y);
                        imagesetpixel($rotate, $y, $dest_y - $x - 1, $color);
                     }
                  }
                  break;

               case 180:
                  for ($y = 0; $y < ($src_y); $y++) {
                     for ($x = 0; $x < ($src_x); $x++) {
                        $color = imagecolorat($src_img, $x, $y);
                        imagesetpixel($rotate, $dest_x - $x - 1, $dest_y - $y - 1, $color);
                     }
                  }
                  break;

               default:
                  $rotate = $src_img;
                  break;
            };

            return $rotate;

         }

         $img = imagerotate($img, $degrees);

      }

      return $img;

   } // rotateImage()

   /**
    * return all assigned tags for the specified photo
    */
   private function get_photo_tags($idx)
   {
      $result = $this->db->db_query("
         SELECT t.id, t.name
         FROM tags t
         INNER JOIN photo_tags pt
            ON t.id=pt.tag_id
         WHERE pt.photo_id='". $idx ."'
      ");

      $tags = Array();

      while($row = $this->db->db_fetch_object($result))
         $tags[$row['id']] = $row['name'];

      return $tags;

   } // get_photo_tags()

   /**
    * create on-the-fly images with text within
    */
   public function showTextImage($txt, $color=000000, $space=4, $font=4, $w=300)
   {
      if (strlen($color) != 6) 
         $color = 000000;

      $int = hexdec($color);
      $h = imagefontheight($font);
      $fw = imagefontwidth($font);
      $txt = explode("\n", wordwrap($txt, ($w / $fw), "\n"));
      $lines = count($txt);
      $im = imagecreate($w, (($h * $lines) + ($lines * $space)));
      $bg = imagecolorallocate($im, 255, 255, 255);
      $color = imagecolorallocate($im, 0xFF & ($int >> 0x10), 0xFF & ($int >> 0x8), 0xFF & $int);
      $y = 0;

      foreach ($txt as $text) {
         $x = (($w - ($fw * strlen($text))) / 2);
         imagestring($im, $font, $x, $y, $text, $color);
         $y += ($h + $space);
      }

      Header("Content-type: image/png");
      ImagePng($im);

   } // showTextImage()

   /**
    * check if all requirements are met
    */
   private function checkRequirements()
   {
      if(!function_exists("imagecreatefromjpeg")) {
         print "PHP GD library extension is missing<br />\n";
         $missing = true;
      }

      if(!function_exists("sqlite3_open")) {
         print "PHP SQLite3 library extension is missing<br />\n";
         $missing = true;
      }

      /* Check for HTML_AJAX PEAR package, lent from Horde project */
      ini_set('track_errors', 1);
      @include_once 'HTML/AJAX/Server.php';
      if(isset($php_errormsg) && preg_match('/Failed opening.*for inclusion/i', $php_errormsg)) {
         print "PEAR HTML_AJAX package is missing<br />\n";
         $missing = true;
      }
      @include_once 'Calendar/Calendar.php';
      if(isset($php_errormsg) && preg_match('/Failed opening.*for inclusion/i', $php_errormsg)) {
         print "PEAR Calendar package is missing<br />\n";
         $missing = true;
      }
      ini_restore('track_errors');

      if(isset($missing))
         return false;

      return true;

   } // checkRequirements()

   private function _debug($text)
   {
      if($this->fromcmd) {
         print $text;
      }

   } // _debug()

   /**
    * check if specified MIME type is supported
    */
   public function checkifImageSupported($mime)
   {
      if(in_array($mime, Array("image/jpeg")))
         return true;

      return false;

   } // checkifImageSupported()

   public function _warning($text)
   {
      print "<img src=\"resources/green_info.png\" alt=\"warning\" />\n";
      print $text;

   } // _warning()

   /**
    * output calendard input fields
    */
   private function get_calendar($mode)
   {
      $year = $_SESSION[$mode .'_date'] ? date("Y", strtotime($_SESSION[$mode .'_date'])) : date("Y");
      $month = $_SESSION[$mode .'_date'] ? date("m", strtotime($_SESSION[$mode .'_date'])) : date("m");
      $day = $_SESSION[$mode .'_date'] ? date("d", strtotime($_SESSION[$mode .'_date'])) : date("d");

      $output = "<input type=\"text\" size=\"3\" id=\"". $mode ."year\" value=\"". $year ."\"";
      if(!isset($_SESSION[$mode .'_date'])) $output.= " disabled=\"disabled\"";
      $output.= " />\n";
      $output.= "<input type=\"text\" size=\"1\" id=\"". $mode ."month\" value=\"". $month ."\"";
      if(!isset($_SESSION[$mode .'_date'])) $output.= " disabled=\"disabled\"";
      $output.= " />\n";
      $output.= "<input type=\"text\" size=\"1\" id=\"". $mode ."day\" value=\"". $day ."\"";
      if(!isset($_SESSION[$mode .'_date'])) $output.= " disabled=\"disabled\"";
      $output.= " />\n";
      return $output;

   } // get_calendar()

   /**
    * output calendar matrix
    */
   public function get_calendar_matrix($year = 0, $month = 0, $day = 0)
   {
      if (!isset($year)) $year = date('Y');
      if (!isset($month)) $month = date('m');
      if (!isset($day)) $day = date('d');
      $rows = 1;
      $cols = 1;
      $matrix = Array();

      require_once CALENDAR_ROOT.'Month/Weekdays.php';
      require_once CALENDAR_ROOT.'Day.php';

      // Build the month
      $month = new Calendar_Month_Weekdays($year,$month);

      // Create links
      $prevStamp = $month->prevMonth(true);
      $prev = "javascript:setMonth(". date('Y',$prevStamp) .", ". date('n',$prevStamp) .", ". date('j',$prevStamp) .");";
      $nextStamp = $month->nextMonth(true);
      $next = "javascript:setMonth(". date('Y',$nextStamp) .", ". date('n',$nextStamp) .", ". date('j',$nextStamp) .");";

      $selectedDays = array (
         new Calendar_Day($year,$month,$day),
         new Calendar_Day($year,12,25),
      );

      // Build the days in the month
      $month->build($selectedDays);

      $this->tmpl->assign('current_month', date('F Y',$month->getTimeStamp()));
      $this->tmpl->assign('prev_month', $prev);
      $this->tmpl->assign('next_month', $next);

      while ( $day = $month->fetch() ) {
   
         if(!isset($matrix[$rows]))
            $matrix[$rows] = Array();

         $string = "";

         $dayStamp = $day->thisDay(true);
         $link = "javascript:setCalendarDate(". date('Y',$dayStamp) .", ". date('n',$dayStamp).", ". date('j',$dayStamp) .");";

         // isFirst() to find start of week
         if ( $day->isFirst() )
            $string.= "<tr>\n";

         if ( $day->isSelected() ) {
            $string.= "<td class=\"selected\">".$day->thisDay()."</td>\n";
         } else if ( $day->isEmpty() ) {
            $string.= "<td>&nbsp;</td>\n";
         } else {
            $string.= "<td><a class=\"calendar\" href=\"".$link."\">".$day->thisDay()."</a></td>\n";
         }

         // isLast() to find end of week
         if ( $day->isLast() )
            $string.= "</tr>\n";

         $matrix[$rows][$cols] = $string;

         $cols++;

         if($cols > 7) {
            $cols = 1;
            $rows++;
         }
      }

      $this->tmpl->assign('matrix', $matrix);
      $this->tmpl->assign('rows', $rows);
      $this->tmpl->show("calendar.tpl");

   } // get_calendar_matrix()

   /**
    * output export page
    */
   public function getExport($mode)
   {
      $pictures = $this->getPhotoSelection();
      $current_tags = $this->getCurrentTags();  

      if(!isset($_SERVER['HTTPS'])) $protocol = "http";
      else $protocol = "https";

      $server_name = $_SERVER['SERVER_NAME'];

      foreach($pictures as $picture) {

         $orig_url = $protocol ."://". $server_name . $this->cfg->web_path ."index.php?mode=showp&id=". $picture;
         if($current_tags != "") {
            $orig_url.= "&tags=". $current_tags;
         } 
         if(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
            $orig_url.= "&from_date=". $_SESSION['from_date'] ."&to_date=". $_SESSION['to_date'];
         }

         $thumb_url = $protocol ."://". $server_name . $this->cfg->web_path ."phpfspot_img.php?idx=". $picture ."&width=". $this->cfg->thumb_width;

         switch($mode) {

            case 'HTML':
               // <a href="%pictureurl%"><img src="%thumbnailurl%" ></a>
               print htmlspecialchars("<a href=\"". $orig_url ."\"><img src=\"". $thumb_url ."\" /></a>") ."<br />\n";
               break;
               
            case 'MoinMoin':
               // [%pictureurl% %thumbnailurl%]
               print htmlspecialchars(" * [".$orig_url." ".$thumb_url."&fake=1.jpg]") ."<br />\n";
               break;
         }

      }

   } // getExport()

   /**
    * return all selected tags as one string
    */
   private function getCurrentTags()
   {
      $current_tags = "";
      if($_SESSION['selected_tags'] != "") {
         foreach($_SESSION['selected_tags'] as $tag)
            $current_tags.= $tag .",";
         $current_tags = substr($current_tags, 0, strlen($current_tags)-1);
      }
      return $current_tags;

   } // getCurrentTags()

   /**
    * return the current photo
    */
   public function getCurrentPhoto()
   {
      if(isset($_SESSION['current_photo'])) {
         print $_SESSION['current_photo'];
      }
   } // getCurrentPhoto()

   /**
    * tells the client browser what to do
    *
    * this function is getting called via AJAX by the
    * client browsers. it will tell them what they have
    * to do next. This is necessary for directly jumping
    * into photo index or single photo view when the are
    * requested with specific URLs
    */
   public function whatToDo()
   {
      if(isset($_SESSION['selected_tags']) && !empty($_SESSION['selected_tags'])) {
         return "showpi_tags";
      }
      elseif(isset($_SESSION['from_date']) && isset($_SESSION['to_date'])) {
         return "showpi_date";
      }
      elseif(isset($_SESSION['current_photo']) && $_SESSION['start_action'] == 'showp') {
         return "show_photo";
      }
      elseif(isset($_SESSION['start_action']) && $_SESSION['start_action'] == 'showpi') {
         return "showpi";
      }

      return "nothing special";

   } // whatToDo()

   /**
    * return the current process-user
    */
   private function getuid()
   {
      if($uid = posix_getuid()) {
         if($user = posix_getpwuid($uid)) {
            return $user['name'];
         }
      }
   
      return 'n/a';
   
   } // getuid()

}

?>
